SystemOrganization addCategory: #GraphicsExternal!
SystemOrganization addCategory: #'GraphicsExternal-Tests'!

DisplayScreen subclass: #ExternalScreen
	instanceVariableNames: 'argbMap allocatedForms'
	classVariableNames: ''
	poolDictionaries: ''
	category: 'GraphicsExternal'!

!ExternalScreen commentStamp: '<historical>' prior: 0!
I represent a DisplayScreen that is not part of the Squeak universe. Subclasses must implement the appropriate primitives for creating, destroying and allocating the appropriate external objects.

Note: It is assumed that all external display surfaces are accessible by FXBlt, meaning that any support code must register the surfaces with the surface plugin. This requires that the support code will have a way of accessing the bits of the surface. Although this can be terribly expensive (such as on X where a roundtrip to the server might be required or for an OpenGL display where glReadPixels usually is slow as hell) the appropriate methods should be implemented. This allows for a gradual transition to less expensive model (such as implementing an X11Canvas supporting the drawing primitives of X) and is therefore the preferred solution.

In the eventual case that it's known that BitBlt/FXBlt will *never* be used in conjunction with a particular drawing surface, the support code should return a handle that is a) not a SmallInteger (these are used by the surface plugin) and b) not of the 'bitsSize' of a Form. One possible representation for such a handle would be a ByteArray of a non-integral word size (e.g., a ByteArray of size 5,6, or 7). In this case, all attempts to use FXBlt with the drawing surface will simply fail.
!

!ExternalScreen methodsFor: 'form support' stamp: 'ar 5/28/2000 16:55'!
allocateForm: extentPoint
	"Allocate a new form which is similar to the receiver and can be used for accelerated blts"
	| formHandle displayForm |
	formHandle := self primAllocateForm: self depth width: extentPoint x height: extentPoint y.
	formHandle = nil ifTrue:[^super allocateForm: extentPoint].
	displayForm := (ExternalForm extent: extentPoint depth: self depth bits: nil) 
		setExternalHandle: formHandle on: self.
	allocatedForms at: displayForm put: displayForm executor.
	^displayForm! !

!ExternalScreen methodsFor: 'color mapping' stamp: 'cbc 4/24/2017 21:35'!
balancedPatternFor: aColor
	"Return the pixel word for representing the given color on the receiver"
	^self bitPatternFor: aColor! !

!ExternalScreen methodsFor: 'color mapping' stamp: 'cbc 4/24/2017 21:40'!
bitPatternFor: aColor
	"Return the pixel word for representing the given color on the receiver"
	^aColor isColor
		ifTrue: [Bitmap with: (self pixelWordFor: aColor)]
		ifFalse: [aColor bitPatternForDepth: self depth]! !

!ExternalScreen methodsFor: 'accessing' stamp: 'ar 5/27/2000 20:18'!
colormapFromARGB
	"Return a ColorMap mapping from canonical ARGB pixel values into the receiver"
	^argbMap ifNil:[argbMap := ColorMap mappingFromARGB: self rgbaBitMasks].! !

!ExternalScreen methodsFor: 'color mapping' stamp: 'cbc 4/24/2017 21:44'!
colormapToARGB
	"Return a ColorMap mapping from the receiver into canonical ARGB space."
	^self colormapFromARGB inverseMap! !

!ExternalScreen methodsFor: 'blitting support' stamp: 'ar 5/28/2000 16:16'!
copyBits: sourceRect from: sourceForm at: destOrigin clippingBox: clipRect rule: rule fillColor: hf map: map
	"Attempt to accelerate blts to the receiver"
	| r |
	((self isBltAccelerated: rule for: sourceForm) and:[map == nil and:[hf == nil]]) ifTrue:[
		"Try an accelerated blt"
		r := (destOrigin extent: sourceRect extent) intersect: (clipRect intersect: clippingBox).
		r area <= 0 ifTrue:[^self].
		(self primBltFast: bits from: sourceForm getExternalHandle
			at: r origin from: sourceRect origin
			extent: r extent) ifNotNil:[^self].
	].
	^super copyBits: sourceRect from: sourceForm at: destOrigin clippingBox: clipRect rule: rule fillColor: hf map: map! !

!ExternalScreen methodsFor: 'initialize-release' stamp: 'ar 5/28/2000 18:44'!
destroy
	"Destroy the receiver"
	allocatedForms ifNotNil:[
		allocatedForms lock. "Make sure we don't get interrupted"
		allocatedForms forceFinalization. "Clean up all lost references"
		allocatedForms keys do:[:stillValid| stillValid shutDown].
		"All remaining references are simply destroyed"
		allocatedForms associationsDo:[:assoc| assoc key: nil].
		allocatedForms forceFinalization. "destroy all others"
		allocatedForms := nil.
	].
	bits ifNotNil:[self primDestroyDisplaySurface: bits].
	bits := nil.! !

!ExternalScreen methodsFor: 'form support' stamp: 'ar 5/28/2000 15:56'!
destroyForm: anExternalForm
	"Destroy the given external form"
	self primDestroyForm: anExternalForm getExternalHandle.
	anExternalForm setExternalHandle: nil on: nil.
	allocatedForms removeKey: anExternalForm ifAbsent:[].! !

!ExternalScreen methodsFor: 'blitting support' stamp: 'ar 5/28/2000 16:16'!
displayOn: destForm at: destOrigin clippingBox: clipRect rule: rule fillColor: hf
	"Attempt to accelerate blts to aDisplayMedium"
	| sourceRect |
	((self isBltAccelerated: rule for: destForm) and:[hf = nil]) ifTrue:[
		"Try an accelerated blt"
		sourceRect := (clipRect translateBy: destOrigin negated) intersect: clippingBox.
		(self primBltFast: bits to: destForm getExternalHandle
			at: 0@0 from: sourceRect origin
			extent: sourceRect extent) ifNotNil:[^self]].
	destForm copyBits: self boundingBox
		from: self
		at: destOrigin + self offset
		clippingBox: clipRect
		rule: rule
		fillColor: hf
		map: (self colormapIfNeededFor: destForm).
! !

!ExternalScreen methodsFor: 'blitting support' stamp: 'ar 5/27/2000 21:09'!
fill: aRectangle rule: anInteger fillColor: aColor 
	"Replace a rectangular area of the receiver with the pattern described by aForm 
	according to the rule anInteger."
	| rect |
	(self isFillAccelerated: anInteger for: aColor) ifTrue:[
		rect := aRectangle intersect: clippingBox.
		(self primFill: bits
			color: (self pixelWordFor: aColor)
			x: rect left
			y: rect top
			w: rect width
			h: rect height) ifNotNil:[^self]].
	^super fill: aRectangle rule: anInteger fillColor: aColor! !

!ExternalScreen methodsFor: 'initialize-release' stamp: 'ar 5/27/2000 01:30'!
finish
	"Flush the receiver"
	self primFinish: bits.
	"Now is the time to do some cleanup"
	allocatedForms unlock.
	allocatedForms finalizeValues.! !

!ExternalScreen methodsFor: 'initialize-release' stamp: 'ar 5/27/2000 01:30'!
flush
	"Flush the receiver"
	self primFlush: bits.! !

!ExternalScreen methodsFor: 'testing' stamp: 'ar 5/15/2001 16:15'!
hasNonStandardPalette
	"Quite possible."
	^true! !

!ExternalScreen methodsFor: 'blitting support' stamp: 'ar 5/28/2000 15:52'!
isBltAccelerated: ruleInteger for: aForm
	"Return true if the receiver can perform accelerated blt operations by itself.
	It is assumed that blts of forms allocated on the receiverusing Form>>over 
	may be accelerated.
	Although some hardware may allow source-key blts (that is, Form>>paint or similar)
	this is usually questionable and the additional effort for allocating and
	maintaining the OS form doesn't quite seem worth the effort."
	^aForm displayScreen == self and:[ruleInteger = Form over]! !

!ExternalScreen methodsFor: 'testing' stamp: 'ar 5/27/2000 20:17'!
isExternalForm
	"Sorta. Kinda."
	^true! !

!ExternalScreen methodsFor: 'blitting support' stamp: 'ar 5/27/2000 16:09'!
isFillAccelerated: ruleInteger for: aColor
	"Return true if the receiver can perform accelerated fill operations by itself.
	It is assumed that the hardware can accelerate plain color fill operations."
	^ruleInteger = Form over and:[aColor isColor]! !

!ExternalScreen methodsFor: 'color mapping' stamp: 'cbc 4/24/2017 21:46'!
pixelValueFor: aColor
	"Return the pixel word for representing the given color on the receiver"
	^self colormapFromARGB mapPixel: (aColor pixelValueForDepth: 32)! !

!ExternalScreen methodsFor: 'color mapping' stamp: 'cbc 4/24/2017 21:48'!
pixelWordFor: aColor
	"Return the pixel word for representing the given color on the receiver"
	| basicPattern |
	basicPattern := self pixelValueFor: aColor.
	self depth = 32 
		ifTrue:[^basicPattern]
		ifFalse:[^aColor pixelWordFor: self depth filledWith: basicPattern]! !

!ExternalScreen methodsFor: 'primitives-forms' stamp: 'ar 5/27/2000 17:22'!
primAllocateForm: d width: w height: h
	"Primitive. Allocate a form with the given parameters"
	^nil! !

!ExternalScreen methodsFor: 'primitives-display' stamp: 'ar 5/27/2000 17:18'!
primBltFast: displayHandle from: sourceHandle at: destOrigin from: sourceOrigin extent: extent
	"Primitive. Perform a fast blt operation. Return the receiver if successful."
	^nil! !

!ExternalScreen methodsFor: 'primitives-display' stamp: 'ar 5/28/2000 01:46'!
primBltFast: displayHandle to: dstHandle at: destOrigin from: sourceOrigin extent: extent
	"Primitive. Perform a fast blt operation. Return the receiver if successful."
	^nil! !

!ExternalScreen methodsFor: 'primitives-display' stamp: 'ar 5/27/2000 17:18'!
primCreateDisplaySurface: d width: w height: h
	"Primitive. Create a new external display surface. Return the handle used to identify the receiver. Fail if the surface cannot be created."
	^nil! !

!ExternalScreen methodsFor: 'primitives-display' stamp: 'ar 5/27/2000 17:18'!
primDestroyDisplaySurface: aHandle
	"Primitive. Destroy the display surface associated with the given handle."
	^nil! !

!ExternalScreen methodsFor: 'primitives-forms' stamp: 'ar 5/27/2000 17:22'!
primDestroyForm: aHandle
	"Primitive. Destroy the form associated with the given handle."
	^nil! !

!ExternalScreen methodsFor: 'primitives-display' stamp: 'ar 5/27/2000 17:19'!
primDisplay: aHandle colorMasksInto: anArray
	"Primitive. Store the bit masks for each color into the given array."
	^nil! !

!ExternalScreen methodsFor: 'primitives-display' stamp: 'ar 5/27/2000 17:19'!
primFill: handle color: pixelWord x: x y: y w: w h: h
	"Primitive. Perform an accelerated fill operation on the receiver."
	^nil! !

!ExternalScreen methodsFor: 'primitives-display' stamp: 'ar 5/27/2000 17:20'!
primFinish: aHandle
	"Primitive. Finish all rendering operations on the receiver.
	Do not return before all rendering operations have taken effect."
	^nil! !

!ExternalScreen methodsFor: 'primitives-display' stamp: 'ar 5/27/2000 17:21'!
primFlush: aHandle
	"Primitive. If any rendering operations are pending, force them to be executed.
	Do not wait until they have taken effect."
	^nil! !

!ExternalScreen methodsFor: 'primitives-forms' stamp: 'ar 5/27/2000 17:22'!
primForm: aHandle colorMasksInto: anArray
	"Primitive. Store the bit masks for each color into the given array."
	^nil! !

!ExternalScreen methodsFor: 'initialize-release' stamp: 'ar 5/27/2000 20:16'!
release
	"I am no longer Display. Release any resources if necessary"
	self destroy! !

!ExternalScreen methodsFor: 'accessing' stamp: 'ar 5/27/2000 20:18'!
rgbaBitMasks
	"Return the masks for specifying the R,G,B, and A components in the receiver"
	| rgbaBitMasks |
	rgbaBitMasks := Array new: 4.
	self primDisplay: bits colorMasksInto: rgbaBitMasks.
	^rgbaBitMasks! !

!ExternalScreen methodsFor: 'form support' stamp: 'ar 5/27/2000 18:03'!
rgbaBitMasksOfForm: anExternalForm
	| rgbaBitMasks |
	rgbaBitMasks := Array new: 4.
	self primForm: anExternalForm getExternalHandle colorMasksInto: rgbaBitMasks.
	^rgbaBitMasks! !

!ExternalScreen methodsFor: 'private' stamp: 'ar 5/28/2000 19:00'!
setExtent: aPoint depth: bitsPerPixel
	"Create a 3D accelerated display screen"
	| screen |
	(bits isInteger and:[depth == bitsPerPixel and: [aPoint = self extent and: 
					[self supportsDisplayDepth: bitsPerPixel]]]) ifFalse: [
		bits ifNotNil:[self primDestroyDisplaySurface: bits].
		bits := nil.  "Free up old bitmap in case space is low"
		DisplayChangeSignature := (DisplayChangeSignature ifNil: [0]) + 1.
		(self supportsDisplayDepth: bitsPerPixel)
			ifTrue:[depth := bitsPerPixel]
			ifFalse:["Search for a suitable depth"
					depth := self findAnyDisplayDepthIfNone:[nil]].
		depth == nil ifFalse:[
			bits := self primCreateDisplaySurface: depth 
					width: aPoint x height: aPoint y].
		"Bail out if surface could not be created"
		(bits == nil) ifTrue:[
			screen := DisplayScreen extent: aPoint depth: bitsPerPixel.
			self == Display ifTrue:[
				Display := screen.
				Display beDisplay].
			^screen].
		width := aPoint x.
		height := aPoint y.
	].
	clippingBox := super boundingBox.
	allocatedForms ifNil:[
		allocatedForms := ExternalFormRegistry new.
		WeakArray addWeakDependent: allocatedForms].
! !

!ExternalScreen methodsFor: 'initialize-release' stamp: 'ar 5/28/2000 19:00'!
shutDown 
	"Minimize Display memory saved in image"
	self destroy.
	width := 240.
	height := 120.
	bits := Bitmap new: self bitsSize.! !

!ExternalScreen methodsFor: 'primitives-display' stamp: 'ar 5/27/2000 17:21'!
supportsDisplayDepth: pixelDepth
	"Return true if this pixel depth is supported on the current host platform."
	^false! !

TestCase subclass: #EtoysExternalFormTest
	instanceVariableNames: ''
	classVariableNames: ''
	poolDictionaries: ''
	category: 'GraphicsExternal-Tests'!

!EtoysExternalFormTest methodsFor: 'as yet unclassified' stamp: 'cbc 4/17/2017 09:38'!
testBlitToAndFromExternalForm
	"Ensure that we can blit to/from all 4 permutatations of Form and ExternalForm."
	| source external1 external2 destination |
	source := Cursor wait asCursorForm asFormOfDepth: 32.
	destination := Form extent: source extent depth: 32.
	external1 := ExternalForm extent: source extent depth: 32.
	external2 := ExternalForm extent: source extent depth: 32.
	self shouldnt: [source bits = destination bits].
	source displayOn: external1.
	external1 displayOn: external2.
	external2 displayOn: destination.
	self should: [source bits = destination bits].! !

Form subclass: #ExternalForm
	instanceVariableNames: 'display argbMap'
	classVariableNames: ''
	poolDictionaries: ''
	category: 'GraphicsExternal'!

!ExternalForm commentStamp: '<historical>' prior: 0!
An ExternalForm is just like any other form. It's only difference is that it is allocated on a specific display and can be used for accelerated blts on the particular display.

Upon shutdown of the system ExternalForms will be deallocated from the display and be kept in their internalized form.!

!ExternalForm methodsFor: 'color mapping' stamp: 'cbc 4/24/2017 21:35'!
balancedPatternFor: aColor
	"Return the pixel word for representing the given color on the receiver"
	self hasNonStandardPalette
		ifTrue:[^self bitPatternFor: aColor]
		ifFalse:[^super balancedPatternFor: aColor]! !

!ExternalForm methodsFor: 'color mapping' stamp: 'cbc 4/24/2017 21:39'!
bitPatternFor: aColor
	"Return the pixel word for representing the given color on the receiver"
	(aColor isColor and: [self hasNonStandardPalette]) ifTrue: [^Bitmap with: (self pixelWordFor: aColor)].
	^aColor bitPatternForDepth: self depth! !

!ExternalForm methodsFor: 'accessing' stamp: 'ar 5/27/2000 20:13'!
colormapFromARGB
	"Return a ColorMap mapping from canonical ARGB pixel values into the receiver"
	^argbMap ifNil:[argbMap := ColorMap mappingFromARGB: self rgbaBitMasks].! !

!ExternalForm methodsFor: 'accessing' stamp: 'ar 5/14/2001 21:37'!
colormapFromARGB: aMap
	"Set the ColorMap mapping from canonical ARGB pixel values into the receiver"
	argbMap := aMap! !

!ExternalForm methodsFor: 'color mapping' stamp: 'cbc 4/24/2017 21:46'!
colormapToARGB
	"Return a ColorMap mapping from the receiver into canonical ARGB space."
	^self hasNonStandardPalette 
		ifTrue:[self colormapFromARGB inverseMap]
		ifFalse:[super colormapToARGB]
! !

!ExternalForm methodsFor: 'initialize-release' stamp: 'ar 5/28/2000 17:39'!
destroy
	"Destroy the receiver"
	display ifNotNil:[display destroyForm: self]! !

!ExternalForm methodsFor: 'accessing' stamp: 'ar 5/27/2000 16:56'!
displayScreen
	"Return the display screen the receiver is allocated on."
	^display! !

!ExternalForm methodsFor: 'private' stamp: 'ar 5/27/2000 16:36'!
getExternalHandle
	"Private. Return the virtual handle used to represent the receiver"
	^bits! !

!ExternalForm methodsFor: 'testing' stamp: 'ar 5/15/2001 16:15'!
hasNonStandardPalette
	"Quite possible."
	^display notNil or:[argbMap notNil]! !

!ExternalForm methodsFor: 'testing' stamp: 'ar 5/15/2001 16:17'!
isExternalForm
	"I am an external form but only as long as I'm allocated on a display"
	^display notNil! !

!ExternalForm methodsFor: 'color mapping' stamp: 'cbc 4/24/2017 21:46'!
pixelValueFor: aColor
	"Return the pixel word for representing the given color on the receiver"
	^self hasNonStandardPalette
		ifTrue:[self colormapFromARGB mapPixel: (aColor pixelValueForDepth: 32)]
		ifFalse:[super pixelValueFor: aColor]! !

!ExternalForm methodsFor: 'color mapping' stamp: 'cbc 4/24/2017 21:48'!
pixelWordFor: aColor
	"Return the pixel word for representing the given color on the receiver"
	| basicPattern |
	self hasNonStandardPalette ifFalse:[^super pixelWordFor: aColor].
	basicPattern := self pixelValueFor: aColor.
	self depth = 32 
		ifTrue:[^basicPattern]
		ifFalse:[^aColor pixelWordFor: self depth filledWith: basicPattern]! !

!ExternalForm methodsFor: 'accessing' stamp: 'ar 5/28/2000 17:42'!
rgbaBitMasks
	"Return the masks for specifying the R,G,B, and A components in the receiver"
	display 
		ifNil:[^super rgbaBitMasks]
		ifNotNil:[^display rgbaBitMasksOfForm: self]! !

!ExternalForm methodsFor: 'private' stamp: 'ar 5/27/2000 16:52'!
setExternalHandle: aHandle on: aDisplay
	"Initialize the receiver from the given external handle"
	display := aDisplay.
	bits := aHandle.! !

!ExternalForm methodsFor: 'initialize-release' stamp: 'ar 5/28/2000 17:48'!
shutDown
	"System is going down. Internalize my bits and be finished."
	| copy |
	copy := Form extent: self extent depth: self depth.
	self displayOn: copy.
	copy hibernate. "compact bits of copy"
	self destroy. "Release my external handle"
	bits := copy bits. "Now compressed"
	display := nil. "No longer allocated"
	argbMap := nil. "No longer external"! !

!Form methodsFor: '*GraphicsExternal' stamp: 'ar 5/27/2000 16:56'!
displayScreen
	"Return the display screen the receiver is allocated on. 
	Forms in general are Squeak internal and not allocated on any particular display."
	^nil! !

!Form methodsFor: '*GraphicsExternal' stamp: 'ar 5/27/2000 16:54'!
isExternalForm
	^false! !

WeakKeyDictionary subclass: #ExternalFormRegistry
	instanceVariableNames: 'lockFlag'
	classVariableNames: ''
	poolDictionaries: ''
	category: 'GraphicsExternal'!

!ExternalFormRegistry commentStamp: '<historical>' prior: 0!
The ExternalFormRegistry needs to be synchronized with rendering to prevent forms from being destroyed during rendering. Only at certain points (that is after a rendering cycle is completed) the texture registry may be cleaned up.!

!ExternalFormRegistry methodsFor: 'finalization' stamp: 'ar 5/27/2000 00:56'!
finalizeValues
	"This message is sent when an element has gone away."
	lockFlag == true ifTrue:[^self].
	self forceFinalization.! !

!ExternalFormRegistry methodsFor: 'finalization' stamp: 'ar 5/27/2000 00:55'!
forceFinalization
	self associationsDo:[:assoc|
		assoc key isNil ifTrue:[assoc value destroy].
	].
	super finalizeValues.! !

!ExternalFormRegistry methodsFor: 'accessing' stamp: 'ar 5/26/2000 00:40'!
lock
	lockFlag := true! !

!ExternalFormRegistry methodsFor: 'accessing' stamp: 'ar 5/26/2000 00:40'!
unlock
	lockFlag := false.! !
